// Some PSX libraries in 'C' 
// (c) Blade 2001
// SPU section
// TODO:

#include <spu.h>
#include <pitch.h>
//=======================================================
//==== Lots of defines for the SPU =================================
//=======================================================
#define VAGp			0x70474156		// 'magic number' for a VAG file
#define DMA_ICR		((u_long *)0x1F8010F4)
#define DMA_DPCR		((u_long*)0x1F8010f0)
#define DMA_LINEAR		(0x01000201)
#define SPU_DMA_MADR	((u_long *)0x1f8010c0)
#define SPU_DMA_BCR	((u_long *)0x1f8010c4)
#define SPU_DMA_CHCR	((u_long *)0x1f8010c8)
#define SPU_DELAY		((u_short *)0x1F801014)
#define SPU_VOICE_BASE	((u_short *)0x1F801C00)
#define SPU_MVOL_L		((u_short *)0x1F801D80)
#define SPU_MVOL_R		((u_short *)0x1F801D82)
#define SPU_REVERB_L	((u_short *)0x1F801D84)
#define SPU_REVERB_R	((u_short *)0x1F801D86)
#define SPU_KEY_ON1	((u_short *)0x1F801D88)
#define SPU_KEY_ON2	((u_short *)0x1F801D8A)
#define SPU_KEY_OFF1	((u_short *)0x1F801D8C)
#define SPU_KEY_OFF2	((u_short *)0x1F801D8E)
#define SPU_SB_ADDR	((u_short *)0x1F801DA6)
#define SPU_REG0		((u_short *)0x1F801DAA)
#define SPU_REG1		((u_short *)0x1F801DAC)
#define SPU_STATUS		((u_short *)0x1F801DAE)
#define SPU_CDMVOL_L	((u_short *)0x1F801DB0)
#define SPU_CDMVOL_R	((u_short *)0x1F801DB2)
#define SPU_EXTVOL_L	((u_short *)0x1F801DB4)
#define SPU_EXTVOL_R	((u_short *)0x1F801DB6)
#define SPU_KEY_MODEFM1	((u_short *)0x1F801D90)
#define SPU_KEY_MODEFM2	((u_short *)0x1F801D92)
#define SPU_KEY_MODENOISE1	((u_short *)0x1F801D94)
#define SPU_KEY_MODENOISE2	((u_short *)0x1F801D96)
#define SPU_KEY_MODEREVERB1	((u_short *)0x1F801D98)
#define SPU_KEY_MODEREVERB2	((u_short *)0x1F801D9A)
#define SPU_KEY_CHANNELACTIVE1	((u_short *)0x1F801D9C)
#define SPU_KEY_CHANNELACTIVE2	((u_short *)0x1F801D9E)
#define SPU_REVERB_WORKADDR	((u_short *)0x1F801DA2)
#define SPU_REVERBCONFIG		((u_short *)0x1F801DC0)
//-------------------------------------------------------
#define NO_EFFECT		0
#define VIBRATO		1
#define TREMOLO 		2
#define SLIDE			3
#define VOLUME_SLIDE	4
#define PAN			5
//-------------------------------------------------------
#define SLIDE_UP			0x01
#define SLIDE_DOWN			0x02
#define VOICE_VIBRATO		0x04
#define VOICE_TREMOLO		0x07
#define PANNING			0x08
#define SAMPLE_OFFSET		0x09
#define VOL_SLIDE			0x0a
#define PATTERN_JUMP		0x0b
#define SET_VOLUME			0x0c
#define PATTERN_BREAK		0x0d
#define SONG_SPEED			0x0f
//=======================================================
//==== Some internal prototypes ==================================
//=======================================================
void SPU_VoiceOn(u_char voice);
void SPU_VoiceOff(u_char voice);
void SPU_VoiceVol(u_char voice, u_short vol_left, u_short vol_right);
void SPU_VoicePitch(u_char voice, u_short pitch);
void SPU_Voice2Noise(u_char voice);
void SPU_Voice2NoiseMask(u_long mask);
void SPU_Voice2FM(u_char voice);
void SPU_Voice2FMMask(u_long mask);
void SPU_VoiceADSRRaw(u_char voice, u_short adsr1, u_short adsr2);
int printspu (char *s, ...);
int WaitSPU (void);
u_long Mem2Sram (u_char * src_addr, u_long size_in_bytes);
//=======================================================
u_long SRAM_POS;		// start of sound data area of SRAM
u_char FIRST_VOICE;		// first available voice
HITMOD MOD[5];		// mod information
//=======================================================
void SsInit (void)
{
  u_short i;

  *DMA_DPCR |= 0xB0000;		// switch SPU DMA on
  *SPU_MVOL_L = 0x3fff;		// set vol to 50%
  *SPU_MVOL_R = 0x3fff;
  *SPU_REG0 = 0x0000;		// reset the SPU

  WaitSPU ();

  *SPU_REG1 = 0x0004;
  *SPU_REVERB_L = 0x0000;	// set reverb depth off
  *SPU_REVERB_R = 0x0000;
  *SPU_KEY_OFF1 = 0xFFFF;	// switch all notes (voices) off
  *SPU_KEY_OFF2 = 0xFFFF;
  *SPU_KEY_MODEFM1 = 0x0000;	// switch FM mode off
  *SPU_KEY_MODEFM2 = 0x0000;
  *SPU_KEY_MODENOISE1 = 0x0000;
  *SPU_KEY_MODENOISE2 = 0x0000;
  *SPU_KEY_MODEREVERB1 = 0x0000;
  *SPU_KEY_MODEREVERB2 = 0x0000;

  *SPU_CDMVOL_L = 0x0000;	// set CD volume
  *SPU_CDMVOL_R = 0x0000;
  *SPU_EXTVOL_L = 0x0000;	// set EXT volume
  *SPU_EXTVOL_R = 0x0000;

  for (i = 0; i < 24; i++) {
    SPU_VoiceVol (i, 0x0000, 0x0000);	// set channel volumes to 0
    SPU_VoiceADSRRaw (i, 0x80FF, 0xDFFF);
    SPU_VoicePitch (i, 0x1000);		// set channel pitches to 0x1000 (44.1khz)
  }

  SPU_Voice2NoiseMask (0x00000000);	// switch off noise on all channels
  SPU_Voice2FMMask (0x00000000);	// switch off FM on all channels

  WaitSPU ();			// wait again
  *SPU_REG0 = 0x8000;		// switch SPU on
  *SPU_REVERB_WORKADDR = 0x1FFF;	// * 8

  *SPU_REG0 |= 0xC000;		// enable SPU

  FIRST_VOICE = 0;
  SRAM_POS = 0x1010;

}
//=======================================================
void SsStop (void)
{
  *DMA_ICR = (*DMA_ICR & 0x00dfffff) | 0x00800000;
  *SPU_REG0 = 0x0000;
}
//=======================================================
int WaitSPU (void)
{
  int i, j;

  // check if SPU busy
  j = 0;
  while (*SPU_STATUS & 0x7ff) {
    for (i = 0; i < 9216; i++);		// wait a short while
    if (j++ > 200)			// give the SPU many chances
      return (-1);			// 8192 * 255 also gives good results
  }
  return (0);
}
//=======================================================
void SsSetMVol (u_short vol_left, u_short vol_right)
{
  *SPU_MVOL_L = vol_left << 7;
  *SPU_MVOL_R = vol_right << 7;
}
//=======================================================
void SsGetMVol (SndVolume * m_vol)
{
  m_vol->left = *SPU_MVOL_L >> 7;
  m_vol->right = *SPU_MVOL_R >> 7;
}
//=======================================================
void SPU_VoiceOn (u_char voice)
{
  *SPU_KEY_ON1 = (1 << voice) & 0xFFFF;
  *SPU_KEY_ON2 = (1 << voice) >> 16;
}
//=======================================================
void SPU_VoiceOff (u_char voice)
{
  *SPU_KEY_OFF1 = (1 << voice) & 0xFFFF;
  *SPU_KEY_OFF2 = (1 << voice) >> 16;
}
//=======================================================
void SPU_VoiceVol (u_char voice, u_short vol_left, u_short vol_right)
{
  u_short *v_base = (u_short *) (SPU_VOICE_BASE + 8 * voice);
  v_base[0] = vol_left;
  v_base[1] = vol_right;
}
//=======================================================
void SPU_VoicePitch (u_char voice, u_short pitch)
{
  u_short *v_base = (u_short *) (SPU_VOICE_BASE + 8 * voice);
  *(v_base + 2) = pitch;
}
//=======================================================
void SPU_VoiceSetData (u_char voice, u_long sb_addr)
{
  u_short *v_base = (u_short *) (SPU_VOICE_BASE + 8 * voice);
  *(v_base + 3) = sb_addr >> 3;
  *(v_base + 7) = sb_addr >> 3;
}
//=======================================================
void SPU_Voice2Noise (u_char voice)
{
  *SPU_KEY_MODENOISE1 = (1 << voice) & 0xFFFF;
  *SPU_KEY_MODENOISE2 = (1 << voice) >> 16;
}
//=======================================================
void SPU_Voice2NoiseMask (u_long mask)
{
  *SPU_KEY_MODENOISE1 = mask & 0xFFFF;
  *SPU_KEY_MODENOISE2 = mask >> 16;
}
//=======================================================
void SPU_Voice2FM (u_char voice)
{
  *SPU_KEY_MODEFM1 = (1 << voice) & 0xFFFF;
  *SPU_KEY_MODEFM2 = (1 << voice) >> 16;
}
//=======================================================
void SPU_Voice2FMMask (u_long mask)
{
  *SPU_KEY_MODEFM1 = mask & 0xFFFF;
  *SPU_KEY_MODEFM2 = mask >> 16;
}
//=======================================================
void SPU_VoiceADSRRaw (u_char voice, u_short adsr1, u_short adsr2)
{
  u_short *v_base = (u_short *) (SPU_VOICE_BASE + 8 * voice);
  *(v_base + 4) = adsr1;
  *(v_base + 5) = adsr2;
}
//=======================================================
u_long SsGetVoiceMask (void)
{
  return ((*SPU_KEY_CHANNELACTIVE2 << 16) | *SPU_KEY_CHANNELACTIVE1);
}
//=======================================================
long SsVagLoad (VAGHEADER * vag)
{
  u_long i, j;
  u_long size_in_bytes;

  // check it's a valid VAG file
  if (vag->id != VAGp)
    return (-1);

  // datalen is wrong endian, so swap it all around
  i = (vag->datalen << 24) | ((vag->datalen << 8) & 0x00ff0000);
  j = ((vag->datalen >> 8) & 0x0000ff00) | (vag->datalen >> 24);
  size_in_bytes = (i | j);

  vag++;			// move pointer past header

  i = Mem2Sram ((u_char *) vag, size_in_bytes);

  return (i);
}
//=======================================================
u_long Mem2Sram (u_char * src_addr, u_long size_in_bytes)
{
  u_short i;
  u_long sb_addr = SRAM_POS;

  printspu("Loading sound data to SRAM ");

  SRAM_POS += size_in_bytes;
  i = 4 - (SRAM_POS % 4);	// round up to a word alignment
  if (i == 4)
    i = 0;
  SRAM_POS += i;

  // specify destination address in SPU sound buffer
  *SPU_SB_ADDR = (u_short) (sb_addr >> 3);	// destination in SRAM
  WaitSPU ();
  *SPU_REG0 = (*SPU_REG0 & 0xFFCF) | 0x0020;	// DMA write
  WaitSPU ();
  while ((*SPU_REG0 & 0x30) != 0x20);		// wait for transfer ready
  WaitSPU ();
  *SPU_DELAY = (*SPU_DELAY & 0xF0FFFFFF) | 0x20000000;
  WaitSPU ();
  *SPU_DMA_MADR = (u_long)src_addr;		// start address of data in memory

  // specify size of data (extra block if needed)
  if ((size_in_bytes & 0x3F) > 0)
    i = 1;
  else
    i = 0;
  *SPU_DMA_BCR = ((size_in_bytes >> 6) + i) << 16 | 0x10;
  *SPU_DMA_CHCR = DMA_LINEAR;			// start transfer

  // wait for the sample to be transferred and print a progress dot
  for(i = 0; i < (size_in_bytes >> 13); i++) {
    printspu("%c", 254);
    WaitSPU ();
    }
  printspu("\n");

  return(sb_addr);
}
//=======================================================
char SsPlaySound (u_long addr, char voice, int lvol, int rvol)
{
  u_long voice_active_mask;
  u_char i;

  lvol = lvol << 7;
  rvol = rvol << 7;
  if (voice >= 0) {
    SPU_VoiceSetData (voice, addr);
    SPU_VoiceVol (voice, lvol, rvol);
    SPU_VoiceOn (voice);
    return (voice);
  }

  voice_active_mask = ((*SPU_KEY_CHANNELACTIVE2 << 16) | *SPU_KEY_CHANNELACTIVE1);

  for (i = FIRST_VOICE; i < 24; i++)
    if (((voice_active_mask >> i) & 0x00000001) == 1) {
      voice = i;
      break;
    }

  if (voice >= 0) {
    SPU_VoiceSetData (voice, addr);
    SPU_VoiceVol (voice, lvol, rvol);
    SPU_VoiceOn (voice);
  }

  return (voice);
}
//=======================================================
int printspu (char *s, ...)
{
  asm ("
	li $10, 0xA0
	li $9, 0x3F
	jr $10
	");
  return (0);
}
//=======================================================
void PauseMod (u_char num)
{
	int i;
	for (i = 0; i < FIRST_VOICE; i++)
	  SPU_VoiceVol (i, 0x0000, 0x0000);
	MOD[num].on = 0;
}
//=======================================================
void StartMod (u_char num)
{
	MOD[num].on = 1;
}
//=======================================================
void SetModVol (u_char num, u_char mod_vol)
{
	MOD[num].volume = mod_vol;
}
//=======================================================
u_char GetModVol (u_char num)
{
	return(MOD[num].volume);
}
//=======================================================
void ReStartMod (u_char num)
{
  u_char *mod_start = MOD[num].mod_start;
  MOD[num].volume = 23;
  MOD[num].current_row = 0;
  MOD[num].current_order = 0;
  MOD[num].current_pattern = mod_start[952];
  MOD[num].key_on = 0x0000;
  MOD[num].speed = 6;
  MOD[num].tick = 0;
  MOD[num].row_pointer = (u_long *) (mod_start + 1084 + (MOD[num].current_pattern * MOD[num].tracks * 256));
  MOD[num].on = 1;
}
//=======================================================
void InitMod (u_char num, u_char * mod_start)
{
  int i;
  u_char *sram_address;
  u_char highest;
  u_long position;

  MOD[num].mod_start = mod_start;
  MOD[num].volume = 23;
  MOD[num].current_row = 0;
  MOD[num].current_order = 0;
  MOD[num].current_pattern = mod_start[952];
  MOD[num].key_on = 0x0000;
  MOD[num].speed = 6;
  MOD[num].tick = 0;
  MOD[num].song_length = mod_start[950];
  MOD[num].tracks = mod_start[1083] - 48;
  MOD[num].row_pointer = (u_long *) (mod_start + 1084 + (MOD[num].current_pattern * MOD[num].tracks * 256));

  FIRST_VOICE = MOD[num].tracks;
  MOD[num].sample_end = SRAM_POS;

  for (i = 0; i < 30; i++) {
    position = 42 + i * 30;
    MOD[num].sample_address[i] = MOD[num].sample_end;
    MOD[num].sample_end += (mod_start[position] << 8) + mod_start[position + 1];
    MOD[num].sample_volume[i] = mod_start[position + 3];
    }

  // get the number of patterns by looking for the highest one used
  MOD[num].patterns = 0;
  for (i = 0; i < 128; i++) {
    highest = mod_start[952 + i];
  if (highest > MOD[num].patterns)
      MOD[num].patterns = highest;
  }

  MOD[num].pattern_data_size = (MOD[num].patterns + 1) * MOD[num].tracks * 256;
  sram_address = mod_start + 1084 + MOD[num].pattern_data_size;

  for (i = 0; i < MOD[num].tracks; i++) {
    MOD[num].channel_volume[i] = 0;
    MOD[num].last_sample[i] = 0;
    MOD[num].LFO_pointer[i] = 0;
    MOD[num].LFO_speed[i] = 0;
    MOD[num].LFO_depth[i] = 0;
    SPU_VoiceADSRRaw (i, 0x80FF, 0xBFBF);
  }

  MOD[num].sample_start = Mem2Sram (sram_address, MOD[num].sample_end - SRAM_POS + 32);

  MOD[num].on = 1;

}

// Play next row of mod
void PollMod (u_char num)
{
  int i;
  u_char effect_command;
  char effect_data;
  u_char uc;
  u_long w;
  u_long nn;
  u_long offset[8];
  u_short *vwa;
  u_short us1;
  char sc;

  // should we play the MOD?
  if (MOD[num].on == 0)   return;

  // increment the tick
  MOD[num].tick++;

  // update Low Frequency Oscillator pointers
  for (i = 0; i < MOD[num].tracks; i++) {
    MOD[num].LFO_pointer[i] = (MOD[num].LFO_pointer[i] + MOD[num].LFO_speed[i]) % 64;
    }

    // do stuff that needs to be done every tick
    vwa = SPU_VOICE_BASE;

    for (i = 0; i < MOD[num].tracks; i++) {

      switch (MOD[num].effect[i]) {

		case NO_EFFECT:	break;

		case VIBRATO:		sc = modlfo[MOD[num].LFO_pointer[i] + 64 * MOD[num].LFO_depth[i]];
						nn = MOD[num].last_period[i] + sc;
						*(vwa + 2) = pitchtab[nn];
						break;

		case TREMOLO:		sc = modlfo[MOD[num].LFO_pointer[i] + 64 * MOD[num].LFO_depth[i]];
						uc = MOD[num].channel_volume[i] + sc;
						if (uc > 64) uc = 64;
						us1 = (uc * MOD[num].sample_volume[MOD[num].last_sample[i]] * MOD[num].volume) >> 4;
						*vwa = us1;
						*(vwa + 1) = us1;
						break;

		case SLIDE:		MOD[num].last_period[i] += MOD[num].pitch_slide[i];
						nn = MOD[num].last_period[i];
						*(vwa + 2) = pitchtab[nn];
						break;

		case VOLUME_SLIDE:	MOD[num].channel_volume[i] += MOD[num].volume_slide[i];
						if (MOD[num].channel_volume[i] > 64) MOD[num].channel_volume[i] = 64;
						if (MOD[num].channel_volume[i] < 0) MOD[num].channel_volume[i] = 0;
						us1 = (MOD[num].channel_volume[i] *
							MOD[num].sample_volume[MOD[num].last_sample[i]] *
							MOD[num].volume) >> 4;
						*vwa = us1;
						*(vwa + 1) = us1;
						break;

		}
		vwa += 8;
	}

  // we can return unless speed = tick
  if (MOD[num].tick < MOD[num].speed) return;

  MOD[num].key_on = 0;

  vwa = SPU_VOICE_BASE;

  for (i = 0; i < MOD[num].tracks; i++) {
	offset[i] = 0;
	MOD[num].effect[i] = NO_EFFECT;
	// get data word
	w = *MOD[num].row_pointer;

	MOD[num].row_pointer++;

	// parse for samplenum / instrnum
	uc = ((w >> 20) & 0x0f) | (w & 0xf0);
	if (uc > 0) {
		MOD[num].channel_volume[i] = 64;
		MOD[num].last_sample[i] = uc - 1;
		MOD[num].key_on |= 1 << i;
		}

	effect_command = (w >> 16) & 0x0f;
	if (effect_command != 0) {
		effect_data = (w >> 24) & 0xff;

		switch (effect_command) {
		case SONG_SPEED:		if (effect_data < 20) MOD[num].speed = effect_data;
							  else MOD[num].speed = 800 / effect_data;
							break;

		case SLIDE_UP:			MOD[num].pitch_slide[i] = -effect_data;
							MOD[num].effect[i] = SLIDE;
							break;

		case SLIDE_DOWN:		MOD[num].pitch_slide[i] = effect_data;
							MOD[num].effect[i] = SLIDE;
							break;

		case VOICE_VIBRATO:	MOD[num].LFO_speed[i] = (effect_data & 0xf0) >> 4;
							MOD[num].LFO_depth[i] = effect_data & 0x0f;
							MOD[num].effect[i] = VIBRATO;
							break;

		case VOICE_TREMOLO:	MOD[num].LFO_speed[i] = (effect_data & 0xf0) >> 4;
							MOD[num].LFO_depth[i] = effect_data & 0x0f;
							MOD[num].effect[i] = TREMOLO;
							break;

		case PANNING:			MOD[num].effect[i] = PAN;
							break;
						
		case SAMPLE_OFFSET:	offset[i] = (effect_data << 12) / 28;
							offset[i] &= 0xFFFFFFF0;	// clip to 16-byte boundary
							break;

		case VOL_SLIDE:		MOD[num].volume_slide[i] = ((effect_data & 0xf0) >> 4) - (effect_data & 0x0f);
							MOD[num].effect[i] = VOL_SLIDE;
							break;

		case PATTERN_JUMP:		if (effect_data <= MOD[num].song_length) {
							  MOD[num].current_row = 63;
							  MOD[num].current_order = effect_data - 1;
							  }
							break;

		case SET_VOLUME:		if (effect_data > 64) effect_data = 64;
							MOD[num].channel_volume[i] = effect_data;
							effect_data = (MOD[num].channel_volume[i] * 
									MOD[num].sample_volume[MOD[num].last_sample[i]] * 
									MOD[num].volume) >> 4;
							*vwa = effect_data;
							*(vwa + 1) = effect_data;
							break;
		case PATTERN_BREAK:	MOD[num].current_row = 63;
							break;

		}				// end switch for effects

	}				// end effects

    // check if note on and get note last_period
    nn = ((w & 0x0f) << 8) | ((w >> 8) & 0xff);
    if (nn > 0) {
      MOD[num].last_period[i] = nn;	// remember last_period
      us1 = (MOD[num].channel_volume[i] * MOD[num].sample_volume[MOD[num].last_sample[i]] * MOD[num].volume) >> 4;
      *vwa = us1;
      *(vwa + 1) = us1;
      // set pitch from pitch table
      *(vwa + 2) = pitchtab[nn];
      // set sample addr (including "offset" effect)
      us1 = ((MOD[num].sample_address[MOD[num].last_sample[i]] + offset[i]) >> 3) & 0xffff;
      *(vwa + 3) = us1;
      // set keyon for this channel
      MOD[num].key_on |= 1 << i;
    }				// end if note on

    vwa += 8;
  }				// next channel (track)

  // reset ticks
  MOD[num].tick = 0;

  // do voice ons
  *SPU_KEY_ON1 = MOD[num].key_on;

  // inc row counter
  MOD[num].current_row++;

	// check if we are at the end of the pattern
	if (MOD[num].current_row == 64) {
		MOD[num].current_row = 0;
		MOD[num].current_order++;

		// check for end of song
		if (MOD[num].current_order == MOD[num].song_length) {
			MOD[num].current_order = 0;
			}

		MOD[num].current_pattern = *(MOD[num].mod_start + 952 + MOD[num].current_order);

		// set row_pointer
	MOD[num].row_pointer = (u_long *) (MOD[num].mod_start + 1084 + (MOD[num].current_pattern * MOD[num].tracks * 256));
		}

  return;
}
